/* vim:set sw=4 ts=8 fileencoding=cp1251:::WINDOWS-1251[] */
/*
 * Copyright(C) 2006-2013  
 *
 *    , 
 *    .
 *
 *         
 *  ,     ,    , 
 *    : 
 *   1.        
 *           ,   
 *           . 
 *   2.        
 *           ,   
 *              / 
 *        ,   . 
 * 
 *        
 * /   "  "  -  , 
 *    , ,    , 
 *        
 *  .    ,     
 * ,      ,     
 *      ,    /  
 *  ,    ,  Ѩ 
 * ,   , ,   
 *  ,      
 *   (,     , 
 *  ,  ,    - 
 *    ,       
 *  ),         
 *     
 *
 *  ,    , 
 *         
 *   .
 *
 *  -   
 *     .
 */

/*!
 * \file $RCSfile$
 * \version $Revision: 164730 $
 * \date $Date:: 2017-10-17 20:34:29 +0300#$
 * \author $Author: maxdm $
 *
 * \brief      () Solaris
 *
 */

#include <sys/file.h>
#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/types.h>
#include <sys/time.h>
#include <stddef.h>
#include <netinet/in.h>
#include <inttypes.h>
#include <sys/mutex.h>

#ifndef TERMINAL_LIBRARY
//# include <sys/systm.h>
#endif

#include "drtcsp_io.h"
#include "reader/tchar.h"
#include "wincspc_int.h"
#ifndef USE_STD_MM
#include "lfmm.h"
#include "lfmm_int.h"
#endif
#define MAXCONTEXTS 1000
/*
 *        
 *  .   
 *  
 */
#ifdef USE_STD_MM
const size_t KmemMaxSize = 1000000; /*     */
typedef enum tagKmemHeadMagic {
  HeadMagic       = 0x48656164,
  UnusedHeadMagic = 0x55686472
}       KmemHeadMagic;

typedef enum tagKmemTailMagic {
  TailMagic       = 0x5461696c,
  UnusedTailMagic = 0x5574696c
}       KmemTailMagic;

typedef struct tagKmemHead {
  size_t          size;
  KmemHeadMagic   magic;
}       KmemHead;

typedef struct tagKmemTail {
  KmemTailMagic   magic;
  size_t          size;
}       KmemTail;

static inline KmemHead *get_head(void *p) {
  return &(((KmemHead *)p)[-1]);
}
static inline void *get_body(KmemHead *p) {
  return &(p[1]);
}
static inline KmemTail *get_tail(void *p, size_t s) {
  return (KmemTail *)&(((char *)p)[s-sizeof(KmemHead)-sizeof(KmemTail)]);
}

/*
 *        
 */
const size_t KmemAlignment = sizeof(KmemHead *);

/*!
 * \ingroup MemoryManager
 * \brief  
 *
 *   .
 */
static DWORD
stdAllocMemory(LPCPC_MEMORY_ARENA pArena, CPC_SIZE_T Size, DWORD dwMemPoolId,
	       DWORD ThreadId, LPVOID *out)
{
    KmemHead *pH;
    KmemTail *pT;

    UNUSED(dwMemPoolId);

    Size = (Size + KmemAlignment - 1)&~(KmemAlignment-1);
    if (Size > KmemMaxSize) {
	cmn_err(CE_NOTE, MODNAME": %s: %d: Too big memory alloc %x", __func__,
		__LINE__, Size);
	pH = NULL;
	goto exit;
    }
    Size += sizeof(KmemHead)+sizeof(KmemTail);
    pH = (KmemHead *)kmem_alloc (Size, KM_NOSLEEP); 
    			// TODO:    KM_SLEEP   FPU
			// TODO:     CSPConfig->FuncStruct
    if(pH) {
	pH->magic = HeadMagic;
	pH->size = Size;
	pT = get_tail(get_body(pH), Size);
	pT->magic = TailMagic;
	pT->size = ~Size;
    } else {
	cmn_err(CE_NOTE, MODNAME": %s: %d: No memory", __func__, __LINE__);
    }

 exit:
    if(!pH) {
	cmn_err(CE_NOTE, MODNAME": %s: %d: NTE_NO_MEMORY", __func__, __LINE__);
	return NTE_NO_MEMORY;
    }
    //    AccountMem (hHeap, pMem, TRUE, dwMemPoolId);
    *out=get_body(pH);
    return 0;
}

/*!
 * \ingroup MemoryManager
 * \brief  
 *
 *     .
 */
static DWORD
stdFreeMemory(LPCPC_MEMORY_ARENA pArena, VOID *pMem, DWORD dwMemPoolId)
{
    KmemHead *pH;
    KmemTail *pT;
    size_t size;

    UNUSED(dwMemPoolId);
    if (!pMem) {
	cmn_err(CE_NOTE, MODNAME": %s: %d: Bad args", __func__, __LINE__);
	return NTE_FAIL;
    }
    pH = get_head(pMem);
    if(!pH || pH->magic != HeadMagic || pH->size > KmemMaxSize) {
	if(pH->magic == UnusedHeadMagic)
	    cmn_err(CE_NOTE, MODNAME": %s: %d: Possible double free", __func__, __LINE__);
	cmn_err(CE_NOTE, MODNAME": %s: %d: Head corrupt", __func__, __LINE__);
	return NTE_FAIL;
    }
    size = pH->size;
    pT = get_tail(pMem, size);
    if (!pT || pT->magic != TailMagic || pT->size != ~size) {
	cmn_err(CE_NOTE, MODNAME": %s: %d: Tail corrupt", __func__, __LINE__);
	return NTE_FAIL;
    }
    pH->magic = UnusedHeadMagic;
    pH->size <<= 8;
    pT->magic = UnusedTailMagic;
    pT->size <<= 16;

    kmem_free(pH, size);

    return 0;
}

static void
stdValidateMemory(LPCPC_MEMORY_ARENA pArena)
{
    UNUSED(pArena);
}

static void 
stdStatMemory(LPCPC_MEMORY_ARENA pArena, LPCPC_MEMORY_STATS pStats,
	      DWORD dwMemPoolId)
{
    UNUSED (pArena);
    UNUSED (pStats);
    UNUSED (dwMemPoolId);
}

static void
stdDoneMemory(LPCPC_MEMORY_ARENA pArena)
{
    UNUSED(pArena);
}

static DWORD
stdInitMemory(LPCPC_MEMORY_ARENA *pArena, LONG *PoolSizes, DWORD nPools)
{
    static CPC_MEMORY_ARENA Config;

    bzero(&Config, sizeof(Config));
    Config.pValidateMemory = stdValidateMemory;
    Config.pDoneMemory = stdDoneMemory;
    Config.pAllocMemory = stdAllocMemory;
    Config.pFreeMemory = stdFreeMemory;
    Config.pStatMemory = stdStatMemory;

    if(sizeof(KmemHead)%KmemAlignment != 0 ||
       sizeof(KmemHead)&(KmemAlignment - 1) != 0) {
	cmn_err(CE_NOTE, MODNAME": %s: %d: Compilation error", __func__, __LINE__);
	return NTE_FAIL;
    }
    if(sizeof(KmemTail)%KmemAlignment != 0 ||
       sizeof(KmemTail)&(KmemAlignment - 1) != 0) {
	cmn_err(CE_NOTE, MODNAME": %s: %d: Compilation error", __func__, __LINE__);
	return NTE_FAIL;
    }

    *pArena = &Config;

    return S_OK;
}
#else
static DWORD lfmmInitMemory(LPCPC_MEMORY_ARENA * pArena, mem_block * blk, DWORD maxContexts)
{
    CPC_LFMM_CONFIG Config;
    LONG PoolSizes[16];
    DWORD code;

    bzero (PoolSizes, sizeof(PoolSizes));
    PoolSizes [MP_WORK] += 10000 * (250 + maxContexts); 
    PoolSizes [MP_BIG] += 10000 * (250 + maxContexts); 
    PoolSizes [MP_PRIME_M] += 1024 * (400 + maxContexts); 
    PoolSizes [MP_SEC_M] = 1024 * (400 + maxContexts); 
    PoolSizes [MP_WORK_M] += 2 * 1024 * (400 + maxContexts);
    bzero (&Config,sizeof(Config));
    Config.Buffer=NULL;
    Config.Size=0;
    Config.fSMP=TRUE;
    Config.PoolSizes=PoolSizes;
    Config.nPools=16;
    Config.nCPUs=1;
    code=CPCInitMemoryLF(pArena,&Config);
    if (code != NTE_NO_MEMORY)
	return NTE_FAIL;

    Config.Buffer= kmem_zalloc(Config.Size,KM_SLEEP); /* may sleep on request */
    if(!Config.Buffer) {
	cmn_err(CE_NOTE, MODNAME ": %s: %d: Too big memory alloc %x\n",
	       __func__, __LINE__, Config.Size);
	return NTE_NO_MEMORY;
    }
    blk->buffer=Config.Buffer;
    blk->size=Config.Size;
    code=CPCInitMemoryLF(pArena,&Config);
    if (code != S_OK) {
	kmem_free(Config.Buffer,Config.Size);
	blk->buffer=NULL;
	blk->size=0;
	return code;
    }

    return S_OK;
}

static void drtcsp_lfmmDoneMemory(mem_block *blk)
{
    if (blk->buffer && blk->size)
	kmem_free(blk->buffer,blk->size);
}

#endif

static DWORD CPCAPI 
rwlock_init(LPCPC_RWLOCK pSection, DWORD cbSection, LPVOID lpArg)
{
    char cp_c_version_static_assert[(sizeof(krwlock_t) <= sizeof(*pSection))?1:-1];
    UNUSED(cp_c_version_static_assert);
    if (cbSection<sizeof(krwlock_t))
	return NTE_BAD_DATA;
    rw_init((krwlock_t *)pSection, NULL, RW_DRIVER, lpArg);

    return S_OK;
}

static VOID CPCAPI
rwlock_destroy(LPCPC_RWLOCK pSection)
{
    rw_destroy((krwlock_t*)pSection);
}

static VOID CPCAPI
rwlock_wrlock(LPCPC_RWLOCK pSection)
{
    rw_enter((krwlock_t *)pSection, RW_WRITER);
}

static VOID CPCAPI
rwlock_rdlock(LPCPC_RWLOCK pSection)
{
    rw_enter((krwlock_t *)pSection, RW_READER);
}

static VOID CPCAPI
rwlock_unlock(LPCPC_RWLOCK pSection)
{
    rw_exit((krwlock_t *)pSection);
}

/*
 *   " CSP"
 */
char _depends_on[] = "drv/drvcsp";

/*
 * Data
 */
static  void *drtcsp_softstate;

/*
 *  
 */

extern int CPCAPI
prepare_config(LPCPC_CONFIG CSPConfig, mem_block * blk, int use_fastcode, int use_locks)
{
    CSPConfig->cbSize = sizeof(CPC_CONFIG);
    CPCGetDefaultConfig(CSPConfig,NULL);
    if (!use_fastcode) {
	CSPConfig->FuncStruct.UsesFunctions=CPC_FAST_CODE_NO;
	CSPConfig->FuncStruct.cp_kernel_fpu_begin=0;
	CSPConfig->FuncStruct.cp_kernel_fpu_end=0;
	CSPConfig->FuncStruct.UsedMask=0;
    }
    #ifdef USE_STD_MM
	UNUSED(blk);
	if(stdInitMemory(&CSPConfig->pArena, NULL, 0)) {
	    return NTE_NO_MEMORY;
	}
    #else
	if(lfmmInitMemory(&CSPConfig->pArena, blk, MAXCONTEXTS)) {
	    return NTE_NO_MEMORY;
	}
    #endif    
    CSPConfig->logConfig.name=_TEXT(MODNAME);
    if (use_locks) {
	    CSPConfig->lockFuncs.rwlock_init=rwlock_init;
	    CSPConfig->lockFuncs.rwlock_destroy=rwlock_destroy;
	    CSPConfig->lockFuncs.rwlock_wrlock=rwlock_wrlock;
	    CSPConfig->lockFuncs.rwlock_rdlock=rwlock_rdlock;
	    CSPConfig->lockFuncs.rwlock_unlock=rwlock_unlock;
    }
    return S_OK;
}

extern int CPCAPI
remove_config(LPCPC_CONFIG CSPConfig, mem_block * blk)
{
    UNUSED(CSPConfig);
    #ifdef USE_STD_MM
	UNUSED(blk);
    #else
	drtcsp_lfmmDoneMemory(blk);
    #endif  
    return S_OK;
}

static int
drtcsp_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
  int		instance;
  PDRTCSP	pdrtcsp;
  int error=0;

  switch (cmd) {
  case DDI_ATTACH:
    {
      static int once_init = 0; //TODO: ??   

      if (!once_init) {
	once_init = 1;
	if ((error=init_hook()) != 0) {
	  cmn_err(CE_WARN, MODNAME": %s() init_hook failed (%x)", __func__, error);
	  return DDI_FAILURE;
	}
      }
    }

    /*    */
    instance = ddi_get_instance(dip);

    if (ddi_soft_state_zalloc(drtcsp_softstate, instance) != 0) {
      cmn_err(CE_WARN, MODNAME": ddi_soft_state_zalloc() fail.\n");
      return DDI_FAILURE;
    }

    pdrtcsp = ddi_get_soft_state(drtcsp_softstate, instance);
    pdrtcsp->size = sizeof(*pdrtcsp);
    pdrtcsp->dip = dip;
    pdrtcsp->minor=instance;
    pdrtcsp->state = DRTCSP_ATTACH;

    ddi_set_driver_private(dip, (caddr_t)pdrtcsp);

    /*   /devices */
    if(ddi_create_minor_node(dip, ddi_get_name(dip), 
			     S_IFCHR, instance, DDI_PSEUDO, 0) == DDI_FAILURE) {
      ddi_soft_state_free(drtcsp_softstate, instance);
      return DDI_FAILURE;
    }

    /*  ""  */
    ddi_report_dev(dip);
    return DDI_SUCCESS;

  case DDI_SUSPEND:
    return DDI_SUCCESS;

  case DDI_PM_SUSPEND:
    return DDI_SUCCESS;

  default:
    cmn_err(CE_WARN, MODNAME": unknown cmd=%d.\n", cmd);
    return DDI_FAILURE;
  }
}

/*
 *   
 */
/*ARGSUSED*/
static int
drtcsp_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
	       void **result)
{
  int 	error;
  PDRTCSP	pdrtcsp;
  int	instance = getminor((dev_t)arg);

  switch (infocmd) {
  case DDI_INFO_DEVT2DEVINFO:
    pdrtcsp = ddi_get_soft_state(drtcsp_softstate, instance);
    if (pdrtcsp == NULL) {
      *result = NULL;
      error = DDI_FAILURE;
      cmn_err(CE_WARN, MODNAME": %s() no instance=%d", __func__, instance);
    } else {
      *result = pdrtcsp->dip;
      error = DDI_SUCCESS;
    }
    break;
  case DDI_INFO_DEVT2INSTANCE:
    *result = (void *)instance;
    error = DDI_SUCCESS;
    break;
  default:
    *result = NULL;
    error = DDI_FAILURE;
    cmn_err(CE_WARN, MODNAME": %s() unknown infocmd=%d", __func__, infocmd);
  }

  return error;
}

/*
 *  
 */
static int
drtcsp_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
  PDRTCSP	pdrtcsp;
  int		instance;

  switch (cmd) {
  case DDI_DETACH:
    instance = ddi_get_instance(dip);
    pdrtcsp = ddi_get_soft_state(drtcsp_softstate, instance);
    if(detach_hook(pdrtcsp)) return DDI_FAILURE;
    ddi_remove_minor_node(dip, NULL);
    ddi_soft_state_free(drtcsp_softstate, instance);
    return DDI_SUCCESS;

  case DDI_SUSPEND:
  case DDI_PM_SUSPEND:
    return DDI_SUCCESS;

  default:
    cmn_err(CE_WARN, MODNAME": %s() unknown cmd=%d", __func__, cmd);
    return DDI_FAILURE;
  }
}

static PDRTCSP
get_drtcsp(dev_t dev)
{
  int		instance;
  PDRTCSP pdrtcsp;

  instance = getminor(dev);
  pdrtcsp = ddi_get_soft_state(drtcsp_softstate, instance);
  if (!pdrtcsp)
    cmn_err(CE_WARN, MODNAME": %s() inst=%d drtcsp_softstate=0x%x", __func__,
	    instance, drtcsp_softstate);

  return pdrtcsp;
}

/*ARGSUSED*/
static int
drtcsp_open(dev_t *dev, int openflags, int otyp, cred_t *credp)
{
  PDRTCSP	pdrtcsp;
  int retval=0;

  if(otyp != OTYP_CHR) {
    cmn_err(CE_WARN, MODNAME": %s() ctyp!=OTYP_CHR", __func__);
    return EINVAL;
  }
    
  if(!(pdrtcsp=get_drtcsp(*dev))) return ENXIO;

  retval=open_hook(pdrtcsp);

  return retval;
}

/*ARGSUSED*/
static int
drtcsp_close(dev_t dev, int openflags, int otyp, cred_t *credp)
{
  PDRTCSP	pdrtcsp;

  if(otyp != OTYP_CHR) {
    cmn_err(CE_WARN, MODNAME": %s() ctyp!=OTYP_CHR", __func__);
    return EINVAL;
  }

  if(!(pdrtcsp=get_drtcsp(dev))) return ENXIO;

  return close_hook(pdrtcsp);
}

/*ARGSUSED*/
static int
drtcsp_ioctl(dev_t dev, int cmd, intptr_t arg, int flags, cred_t *credp,
	     int *rvalp)
{
  PDRTCSP     pdrtcsp;
  int         error = EFAULT;
  DRTCSP_DRIVER_KEY   tmp;

  if(!(pdrtcsp=get_drtcsp(dev))) return ENXIO;

  if((error=ioctl_hook(pdrtcsp))) return error;

  switch(cmd){
  case DRTCSP_DRIVER_RNG:
    if(ddi_copyin((const void *)arg, (void *)&tmp, sizeof (tmp), flags)){
      cmn_err(CE_WARN, MODNAME": %s() ddi_copyin failed", __func__);
      break;
    }

    if((error=set_driver_rng(pdrtcsp, &tmp))) break;

    if(ddi_copyout((const void *)&tmp, (void *)arg, sizeof (tmp), flags))
      cmn_err(CE_WARN, MODNAME": %s() ddi_copyout failed", __func__);
    
    break;

  case DRTCSP_SESSION_KEY:
    if(ddi_copyin((const void *)arg, (void *)&tmp, sizeof (tmp), flags)){
      cmn_err(CE_WARN, MODNAME": %s() ddi_copyin failed", __func__);
      break;
    }

    if((error=set_session_key(pdrtcsp, &tmp))) break;

    if(ddi_copyout((const void *)&tmp, (void *)arg, sizeof (tmp), flags))
      cmn_err(CE_WARN, MODNAME": %s() ddi_copyout failed", __func__);
    break;

  case DRTCSP_SELFTEST:
    if(ddi_copyin((const void *)arg, (void *)&tmp, sizeof (tmp), flags)){
      cmn_err(CE_WARN, MODNAME": %s() ddi_copyin failed", __func__);
      break;
    }

    if((error=kernel_selftest(pdrtcsp, &tmp))) break;

    if(ddi_copyout((const void *)&tmp, (void *)arg, sizeof (tmp), flags))
      cmn_err(CE_WARN, MODNAME": %s() ddi_copyout failed", __func__);
    
    break;

  default:
    error = ENOTTY;
    break;
  }
  bzero(&tmp, sizeof(tmp));

  return error;
}

int
trans(unsigned char *data, size_t len, int dir, void *arg)
{
  return uiomove(data, len, (dir==FROM_KERNEL)?UIO_READ:UIO_WRITE, (struct uio*)arg); 
}
static int
drtcsp_read(dev_t dev, struct uio *uiop, cred_t *credp)
{
  PDRTCSP	pdrtcsp;
  int resid=uiop->uio_resid, error;

  if(!(pdrtcsp=get_drtcsp(dev))) return ENXIO;

  error=read_hook(pdrtcsp, &resid, (void*)uiop);
  uiop->uio_resid=resid;

  return error;
}

static int
drtcsp_write(dev_t dev, struct uio *uiop, cred_t *credp)
{
  PDRTCSP	pdrtcsp;
  int resid=uiop->uio_resid, error;

  if(!(pdrtcsp=get_drtcsp(dev))) return ENXIO;

  error=write_hook(pdrtcsp, &resid,  (void*)uiop);
  uiop->uio_resid=resid;

  return error;
}

/*
 *  
 */
static struct cb_ops	drtcsp_cb_ops = {
  drtcsp_open,
  drtcsp_close,
  nodev,			/* ""       */
  nodev,			/*  xxx_print	        */
  nodev,			/*  xxx_dump         	*/
  drtcsp_read,
  drtcsp_write,
  drtcsp_ioctl,
  nodev,			/*  xxx_devmap       	*/
  nodev,			/*  xxx_mmap                 */
  nodev,			/*  xxx_segmap               */
  nochpoll,		        /*  xxx_chpoll               */
  ddi_prop_op,
  NULL,		        /*   STREAMS 	*/
  D_NEW | D_MP,		/*     :) */
  CB_REV,
  nodev,
  nodev
};

/*
 *  ""
 */
static struct dev_ops drtcsp_ops = {
  DEVO_REV,		        /* devo_rev */
  0,			        /* devo_refcnt */
  drtcsp_getinfo,		/* devo_getinfo */
  nulldev,		        /* devo_identify */
  nulldev,	                /* devo_probe */
  drtcsp_attach,		/* devo_attach */
  drtcsp_detach,		/* devo_detach */
  nodev,			/* devo_reset */
  &drtcsp_cb_ops,		/* devo_cb_ops */
  NULL,			/* devo_bus_ops */
  NULL			/* devo_power */
};

/*
 *  
 */
extern	struct	mod_ops mod_driverops;
static	struct modldrv modldrv = {
  &mod_driverops,
  "Test driver of CryptoPro CSP",
  &drtcsp_ops
};

static	struct modlinkage modlinkage = {
  MODREV_1,		/* MODREV_1 manual */
  (void *)&modldrv,	/* device driver linkage */
  NULL		/* list terminator */
};

/*
 *  
 */
int
_init(void)
{
  int	error;

  /*     */
  if((error = ddi_soft_state_init(&drtcsp_softstate, sizeof(DRTCSP), DRTCSP_MAX_CHANNEL+1))){
    cmn_err(CE_WARN, MODNAME": %s() ddi_soft_state_init failed", __func__);
    goto ret;
  }

  /*    */
  if((error = mod_install(&modlinkage)) != 0){
    cmn_err(CE_WARN, MODNAME": %s() mod_install failed", __func__);
    ddi_soft_state_fini(&drtcsp_softstate);
    drtcsp_softstate = NULL;
    goto ret;
  }

ret:
  return error;
}

int
_info(struct modinfo *modinfop)
{
    int	error;

    error = mod_info(&modlinkage, modinfop);
    
    return error;
}

/*
 *  
 */
int
_fini(void)
{
  int status;

  /*    ,  ,   */
  if ((status=fini_hook()) != 0 )
  {
    cmn_err(CE_WARN, MODNAME": %s() fini_hook  failed", __func__);
    status=EPERM;
    goto ret;
  }
  if ((status = mod_remove(&modlinkage)) != 0) {
    cmn_err(CE_WARN, MODNAME": %s() mod_remove failed", __func__);
    goto ret;
  }

  /*    */
  ddi_soft_state_fini(&drtcsp_softstate);
  drtcsp_softstate = NULL;
    
ret:
  return status;
}
